package NLObject;

import Interp.Interp;
import java.util.LinkedList;

public class Pat implements NLObject
{
    public static int SIZE = -1;

    char pat[];

    public Pat(String form)
    {
        if(SIZE == -1)
        {
            System.err.println("Pat.SIZE must be set to sculpture size.");
            System.exit(1);
        }

        pat = new char[SIZE];

        // check for empty '' pat
        if(form.equals("''"))
        {
            for(int i=0; i<SIZE; i++)
            {
                pat[i] = '0';
            }

            return;
        }

        if(form.startsWith("'''"))
        {
            String base = form.substring(3, form.length() - 3);
            if(base.equals("")) base = "0";
            repeatExpand(base);
        }
        else if(form.startsWith("''"))
        {
            String base = form.substring(2, form.length() - 2);
            if(base.equals("")) base = "0";
            percentExpand(base);
        }
        else if(form.startsWith("'"))
        {
            String base = form.substring(1, form.length() - 1);
            if(base.equals("")) base = "0";
            literalExpand(base);
        }
        else
        {
            Interp.error("Invalid format for pattern constructor: " + form);
        }
    }
    
    public void repeatExpand(String base)
    {
        int bInd = 0;
        for(int i=0; i<SIZE; i++)
        {
            pat[i] = base.charAt(bInd);
            bInd = (bInd + 1) % base.length();
        }
    }

    public void percentExpand(String base)
    {
        for(int i=0; i<SIZE; i++)
        {
            // calculate what fraction we are through the pattern
            double frac = ((double)i) / ((double)SIZE);

            // figure out which char in base would map to this one
            int bInd = (int)(frac * base.length());

            pat[i] = base.charAt(bInd);
        }
    }

    public void literalExpand(String base)
    {
        int i = 0;

        for(i=0; i<SIZE && i<base.length(); i++)
        {
            pat[i] = base.charAt(i);
        }

        // if base was too short, fill in 0's
        for(; i<SIZE; i++)
        {
            pat[i] = '0';
        }
    }

    public String getType()
    {
        return "pattern";
    }

    public String toString()
    {
        return new String(pat);
    }

    public Pat lshift(int shiftBy)
    {
        if(shiftBy < 0)
        {
            Interp.error("Cannot shift by negative amount ("+shiftBy+").");
            return null;
        }

        char newpat[] = new char[SIZE];

        int i=0;
        for(i=0; (i + shiftBy) < SIZE; i++)
        {
            newpat[i] = pat[i + shiftBy];
        }
        for(; i<SIZE; i++)
        {
            newpat[i] = '0';
        }

        return new Pat("'" + new String(newpat) + "'");
    }

    public Pat rshift(int shiftBy)
    {
        if(shiftBy < 0)
        {
            Interp.error("Cannot shift by negative amount ("+shiftBy+").");
            return null;
        }

        char newpat[] = new char[SIZE];

        int i = SIZE-1;
        for(; (i - shiftBy) >= 0; i--)
        {
            newpat[i] = pat[i - shiftBy];
        }
        for(; i >= 0; i--)
        {
            newpat[i] = '0';
        }

        return new Pat("'" + new String(newpat) + "'");
    }

    public Pat and(Pat p)
    {
        char newpat[] = new char[SIZE];
        
        for(int i=0; i<SIZE; i++)
        {
            if(pat[i] == '1' && p.pat[i] == '1')
                newpat[i] = '1';
            else
                newpat[i] = '0';
        }

        return new Pat("'" + new String(newpat) + "'");
    }

    public Pat xor(Pat p)
    {
        char newpat[] = new char[SIZE];
        
        for(int i=0; i<SIZE; i++)
        {
            if(pat[i] != p.pat[i])
                newpat[i] = '1';
            else
                newpat[i] = '0';
        }

        return new Pat("'" + new String(newpat) + "'");
    }

    public Pat or(Pat p)
    {
        char newpat[] = new char[SIZE];
        
        for(int i=0; i<SIZE; i++)
        {
            if(pat[i] == '1' ||  p.pat[i] == '1')
                newpat[i] = '1';
            else
                newpat[i] = '0';
        }

        return new Pat("'" + new String(newpat) + "'");
    }

    public Pat inv()
    {
        char newpat[] = new char[SIZE];
        
        for(int i=0; i<SIZE; i++)
        {
            if(pat[i] == '1')
                newpat[i] = '0';
            else
                newpat[i] = '1';
        }

        return new Pat("'" + new String(newpat) + "'");
    }

    public NLObject op(String op, NLObject nlo)
    {
        if(op.equals("~") && nlo instanceof NLVoid)
            return inv(); // alt: return xor(new Pat("''1''"));

        if(op.equals("<<") && nlo instanceof Int)
            return lshift(((Int)nlo).i);

        if(op.equals(">>") && nlo instanceof Int)
            return rshift(((Int)nlo).i);

        if(op.equals("&") && nlo instanceof Pat)
            return and((Pat)nlo);

        if(op.equals("^") && nlo instanceof Pat)
            return xor((Pat)nlo);

        if(op.equals("|") && nlo instanceof Pat)
            return or((Pat)nlo);

        if(op.equals("in") && nlo instanceof List)
        {
            String target = toString();
            List hay = (List)nlo;

            for(int i=0; i<hay.list.size(); i++)
            {
                String cand = hay.list.get(i).toString();
                if(target.equals(cand))
                    return new Bool(true);
            }

            return new Bool(false);
        }

        String warn = "Comparing "+ getType() +" and "+ nlo.getType() +".";

        if(op.equals("=="))
        {
            if(nlo instanceof Pat)
            {
                String p1 = toString();
                String p2 = ((Pat)nlo).toString();
                return new Bool(p1.equals(p2));
            }

            Interp.warn(warn);
            return new Bool(false);
        }

        if(op.equals("$") && nlo instanceof List)
        {
            LinkedList l = new LinkedList();
            l.addAll( ((List)nlo).list );
            l.addFirst(nlo);
            return new List(l);
        }

        if(op.equals("!="))
        {
            if(nlo instanceof Pat)
            {
                String p1 = toString();
                String p2 = ((Pat)nlo).toString();
                return new Bool(!p1.equals(p2));
            }

            Interp.warn(warn);
            return new Bool(true);
        }

        String err = "Cannot apply '" + op + "' in '" + getType() +
                     " " + op + " " + nlo.getType() + "'.";
        Interp.error(err);
        return null;
    }
}
